// tcmatch.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "tcmatch.h"

HMODULE hInst;

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			hInst=(HMODULE)hModule;
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

WCHAR* wcslcpy(WCHAR *str1,const WCHAR *str2,int imaxlen)
{
	if ((int)wcslen(str2)>=imaxlen-1) {
		wcsncpy(str1,str2,imaxlen-1);
		str1[imaxlen-1]=0;
	} else
		wcscpy(str1,str2);
	return str1;
}

int countdotsw(WCHAR* buf)
{
	WCHAR* p;
	int result;

	p=buf;
	result=0;
	do {
		p=wcschr(p,'.');
		if (p) {
			p++;
			result++;
		}
	} while (p);
	return result;
}

int hextoint(char ch)
{
	if (ch>='0' && ch<='9')
		return ch-'0';
	else if (ch>='A' && ch<='F')
		return (ch-'A')+10;
	else if (ch>='a' && ch<='f')
		return (ch-'a')+10;
	else
		return 0;
}

WORD PinYinTable[0x9FA5-0x4E00+1];
BOOL tablecreated=false;

void BuildPinYinTable()
{
	DWORD bytesread;
	if (tablecreated)
		return;
	tablecreated=true;
	char name[MAX_PATH],srcname[MAX_PATH];
	memset(&PinYinTable,0,sizeof(PinYinTable));
	GetModuleFileName(hInst,name,sizeof(name)-12);
	char* p=strrchr(name,'\\');
	if (!p)
		return;
	p=strrchr(p,'.');
	if (p)
		p[0]=0;
	strcpy(srcname,name);
	strncat(name,".tbl",MAX_PATH-1);
	strncat(srcname,".txt",MAX_PATH-1);

	HANDLE f=CreateFile(name,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if (f!=INVALID_HANDLE_VALUE) {
		int sz=GetFileSize(f,NULL);
		if (sz==sizeof(PinYinTable)) {    // read binary table!
			ReadFile(f,&PinYinTable,sizeof(PinYinTable),&bytesread,NULL);
			CloseHandle(f);
			return;
		}
		CloseHandle(f);
	}

	// Build table from plain text list!
	// To do this, name the list "Uni2Pinyin". The binary will be created as Uni2Pinyin.bin
	f=CreateFile(srcname,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	int sz=GetFileSize(f,NULL);

	if (f==INVALID_HANDLE_VALUE)
		return;

	char* buf=(char*)malloc(sz+1);
	ReadFile(f,buf,sz,&bytesread,NULL);
	buf[sz]=0;
	CloseHandle(f);

	char* pline=buf;
	char* pend;
	do {
		pend=strchr(pline,0xA);
		if (pend) {
			pend[0]=0;
			pend++;
		}
		if (pline[0]!='#') {
			p=strtok(pline,"\t");
			if (p) {
				WCHAR uchar=0;
				char oldch1=0;
				char oldch2=0;
				if (strlen(p)==4)
					uchar=(hextoint(p[0])<<12)+(hextoint(p[1])<<8)+(hextoint(p[2])<<4)+(hextoint(p[3]));
				if (uchar>=0x4E00 && uchar<=0x9FA5) {
					char* up=p;
					p=strtok(NULL,"\t");
					while (p) {
						char ch=p[0];
						if (ch && ch>='a' && ch<='z') {
							if (!oldch1 && ch) {
								PinYinTable[uchar-0x4E00]=ch-'a'+1;
								oldch1=ch;
							} else if (oldch1 && oldch1!=ch) {
								if (!oldch2) {
									PinYinTable[uchar-0x4E00] |=(ch-'a'+1)<<5;
									oldch2=ch;
								} else if (oldch2 && oldch2!=ch) {
									PinYinTable[uchar-0x4E00] |=(ch-'a'+1)<<10;
									break;
								}
							}
						}
						p=strtok(NULL,"\t");
					}
				}
			}
		}
		pline=pend;
	} while (pline);
	free(buf);
	f=CreateFile(name,GENERIC_WRITE,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,
		CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	if (f!=INVALID_HANDLE_VALUE) {
		WriteFile(f,&PinYinTable,sizeof(PinYinTable),&bytesread,NULL);
		CloseHandle(f);
	}
}

BOOL PinYinMatch(WCHAR Eng,WCHAR Chn)
{
	if (Chn==0x3007)
		return (Eng=='l');
	else if (Chn>=0x4E00 && Chn<=0x9FA5) {
		BuildPinYinTable();
		WORD Dbl=PinYinTable[Chn-0x4E00];
		WCHAR Eng1=(Dbl & 0x1F)-1+'a';
		WCHAR Eng2=((Dbl>>5)&0xF0)-1+'a';
		WCHAR Eng3=((Dbl>>10)&0xF0)-1+'a';
		BOOL result=(Eng==Eng1 || Eng==Eng2 || Eng==Eng3);
		if (!result) {
			// there are only 3 ideographs with more than 3 spellings:
			switch (Chn) {
			case 0x7AD3:	// qian1	fen1	zhi1	yi1	gong1	sheng1
				result=(Eng=='y' || Eng=='g' || Eng=='s');
				break;
			case 0x7AD5:		// shi2		fen1	zhi1	yi1	gong1	sheng1
				result=(Eng=='y' || Eng=='g');
				break;
			case 0x7AE1:		// yi1		gong1	sheng1	bai3	bei4
				result=(Eng=='b');
				break;
			}
		}
		return result;
	} else
		return false;
}

#define wdirtypemax 1023

BOOL __stdcall MatchFileW(WCHAR* Filter,WCHAR* FileName)
{
	BOOL result,failed;
	WCHAR pattern[wdirtypemax+1],buffer[wdirtypemax+1];
	WCHAR *ppat,*pbuf,*pendbuf,*PosOfStar;

	wcslcpy(pattern,Filter,wdirtypemax);
	wcslwr(pattern);
	wcslcpy(buffer,FileName,wdirtypemax);
	wcslwr(buffer);
	result=false;
	failed=false;
	ppat=pattern;
	pbuf=buffer;
	pendbuf=pbuf+wcslen(pbuf);
	PosOfStar=pbuf;
	do {
		if (ppat[0]==0 && pbuf[0]==0)
			result=true;
		else if (ppat[0]=='*') {  //just skip to next
			PosOfStar=pbuf;
			ppat++;
			if (ppat[0]==0)
				result=true;      //* at the end means full match!
		} else {
			if ((ppat[0]=='?' && pbuf[0]) || (ppat[0]==pbuf[0]) || PinYinMatch(ppat[0],pbuf[0])) {  //Match!
				ppat++;
				pbuf++;
			} else if (pbuf[0]==0 && ppat[0]=='.' && ppat[1]=='*' && ppat[2]==0)
				result=true;         //xyz.* matches also xyz
			else if (pbuf[0]==0 && ppat[0]=='.' && ppat[1]==0) {
				// Special case: '.' at end means that pattern and buffer must contain the same number of dots
				ppat[0]=0;  //last pattern-dot doesn't count!
				result=countdotsw(buffer)==countdotsw(pattern);
				failed=!result;
			} else {    //Backtrack!
				while (ppat>pattern && ppat[0]!='*')
					ppat--;
				if (ppat[0]=='*') {
					ppat++;
					PosOfStar++;
					if (PosOfStar>pendbuf)
						failed=true;
					pbuf=PosOfStar;
				} else
					failed=true;
			}
		}
	} while (!result && !failed);
	return result;
}

